@@ -287,6 +287,177 @@ class NewsDataset(Dataset):
         return inputs, label
 
 
+# ==== PATCH: finetune_CLIP (start) ===========================================
+def finetune_CLIP(
+    mode="both",
+    text_path=None,
+    text_column=None,
+    img_dir=None,
+    true_label=None,
+    prompt=None,
+    model_name="best_clip_model.pth",
+    num_epochs=20,
+    batch_size=8,
+    learning_rate=1e-5
+):
+    """
+    Fine-tune CLIP on a text, image, or multimodal dataset.
+
+    Internally remaps labels to 0..N-1 for CrossEntropyLoss.
+    Saves a mapping back to original labels in the checkpoint so predictions can be returned in original space.
+    """
+
+    if mode not in ["text", "image", "both"]:
+        raise ValueError("mode must be one of 'text', 'image', or 'both'")
+
+    use_text  = mode in ["text", "both"]
+    use_image = mode in ["image", "both"]
+
+    if use_text and not text_path:
+        raise ValueError("text_path cannot be empty")
+    if use_image and not img_dir:
+        raise ValueError("img_dir cannot be empty")
+
+    # --- Load dataset ---
+    if text_path.endswith(".csv") or text_path.endswith(".txt"):
+        df = pd.read_csv(text_path)
+    elif text_path.endswith(".jsonl"):
+        df = pd.read_json(text_path, lines=True)
+    elif text_path.endswith((".xlsx", ".xls")):
+        df = pd.read_excel(text_path)
+    else:
+        raise ValueError("Unsupported file format")
+
+    print(f"📂 Loaded {len(df)} records from {os.path.basename(text_path)}")
+
+    if true_label not in df.columns:
+        raise ValueError(f"Label column '{true_label}' not found in dataset")
+
+    # --- Detect unique labels in original space, then remap to 0..N-1 ---
+    unique_labels = sorted(df[true_label].dropna().unique())
+    num_classes   = len(unique_labels)
+    original_to_new = {int(orig): int(idx) for idx, orig in enumerate(unique_labels)}
+    new_to_original = {int(idx): int(orig) for orig, idx in original_to_new.items()}
+
+    # Remap in place for training
+    df[true_label] = df[true_label].map(original_to_new)
+
+    print(f"🔍 Detected {num_classes} classes: {unique_labels}")
+    print(f"🔁 Remapped labels for training -> 0-based indices: {original_to_new}")
+
+    # --- Train/validation split ---
+    if len(df) < 5:
+        train_df, val_df = df, df
+        print("⚠️ Dataset too small for validation split - using full dataset for training and validation.")
+    else:
+        val_size = max(1, int(len(df) * 0.2))
+        train_df = df.iloc[:-val_size].reset_index(drop=True)
+        val_df   = df.iloc[-val_size:].reset_index(drop=True)
+
+    # --- Device selection ---
+    training_device = device
+    if device.type == "mps":
+        training_device = torch.device("cpu")
+        print("⚠️ MPS detected - using CPU to avoid tensor layout issues.")
+    print(f"💻 Using device: {training_device}")
+
+    print("🧠 Training setup:")
+    print(f"   • Mode: {mode}")
+    print(f"   • Text columns: {text_column}")
+    print(f"   • Label column: {true_label} (remapped to 0..{num_classes-1})")
+    print(f"   • Number of classes: {num_classes}")
+    print(f"   • Batch size: {batch_size}, Epochs: {num_epochs}, LR: {learning_rate}")
+    if prompt:
+        print(f"   • Prompt: {prompt}")
+
+    # --- Model & processor ---
+    clip_model = CLIPModel.from_pretrained(base_model)
+    processor  = CLIPProcessor.from_pretrained(base_model)
+
+    train_dataset = NewsDataset(
+        train_df, processor, text_column=text_column, img_dir=img_dir,
+        use_text=use_text, use_image=use_image, true_label=true_label, prompt=prompt
+    )
+    val_dataset = NewsDataset(
+        val_df, processor, text_column=text_column, img_dir=img_dir,
+        use_text=use_text, use_image=use_image, true_label=true_label, prompt=prompt
+    )
+
+    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
+    val_loader   = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)
+
+    # --- Model, optimizer, loss ---
+    model = CLIPClassifier(clip_model, num_classes, use_text, use_image).to(training_device)
+    optimizer = AdamW(model.parameters(), lr=learning_rate)
+    criterion = torch.nn.CrossEntropyLoss()
+
+    # --- Training loop ---
+    best_accuracy = 0.0
+    for epoch in range(num_epochs):
+        model.train()
+        total_loss, correct, total = 0.0, 0, 0
+
+        for _, (inputs, labels) in enumerate(tqdm(train_loader, desc=f"Epoch {epoch+1}/{num_epochs} [Train]")):
+            for k, v in inputs.items():
+                inputs[k] = v.to(training_device)
+            labels = labels.to(training_device)
+
+            optimizer.zero_grad()
+            logits, _ = model(**inputs)
+            loss = criterion(logits, labels)
+            loss.backward()
+            optimizer.step()
+
+            total_loss += float(loss.item())
+            _, predicted = logits.max(1)
+            total += labels.size(0)
+            correct += predicted.eq(labels).sum().item()
+
+        train_acc  = 100.0 * correct / max(1, total)
+        train_loss = total_loss / max(1, len(train_loader))
+
+        # --- Validation ---
+        model.eval()
+        val_correct, val_total, val_loss_accum = 0, 0, 0.0
+        with torch.no_grad():
+            for inputs, labels in tqdm(val_loader, desc=f"Epoch {epoch+1}/{num_epochs} [Val]"):
+                for k, v in inputs.items():
+                    inputs[k] = v.to(training_device)
+                labels = labels.to(training_device)
+
+                logits, _ = model(**inputs)
+                vloss = criterion(logits, labels)
+                val_loss_accum += float(vloss.item())
+                _, predicted = logits.max(1)
+                val_total += labels.size(0)
+                val_correct += predicted.eq(labels).sum().item()
+
+        val_acc  = 100.0 * val_correct / max(1, val_total)
+        val_loss = val_loss_accum / max(1, len(val_loader))
+
+        print(f"Epoch {epoch+1}/{num_epochs}: Train Loss={train_loss:.4f} Acc={train_acc:.2f}% | Val Loss={val_loss:.4f} Acc={val_acc:.2f}%")
+
+        # --- Save best model with label mapping back to original labels ---
+        if val_acc > best_accuracy:
+            best_accuracy = val_acc
+            torch.save({
+                "model_state_dict": model.state_dict(),
+                "epoch": epoch + 1,
+                "best_accuracy": best_accuracy,
+                "num_classes": num_classes,
+                "label_mapping": new_to_original  # key: 0..N-1 -> original label
+            }, model_name)
+            print(f"✅ Model saved - new best validation accuracy: {best_accuracy:.2f}%")
+
+    print(f"🎯 Fine-tuning complete - best validation accuracy: {best_accuracy:.2f}%")
+    if best_accuracy == 0:
+      print("⚠ No valid model saved because validation accuracy stayed at 0%. Check data/labels/training setup.")
+
+    return best_accuracy
+# ==== PATCH: finetune_CLIP (end) =============================================
+
+
+# ==== PATCH: classification_CLIP_finetuned (start) ===========================
 def classification_CLIP_finetuned(
     mode=None,
     text_path=None,
@@ -295,7 +466,7 @@ def classification_CLIP_finetuned(
     prompt=None,
     model_name="best_clip_model.pth",
     batch_size=8,
-    num_classes=None,   # auto-detected
+    num_classes=None,   # auto-detected if None
     predict_column="label",
     true_label=None
 ):
@@ -323,17 +494,25 @@ def classification_CLIP_finetuned(
     try:
         checkpoint = torch.load(model_name, map_location=_device, weights_only=True)
     except Exception:
-        print("⚠️ Safe load failed, retrying with weights_only=False ...")
+        print("⚠️ Safe load failed - retrying with weights_only=False ...")
         checkpoint = torch.load(model_name, map_location=_device, weights_only=False)
 
-    # --- Auto-detect num_classes from checkpoint ---
+    # --- Mapping back to original labels if present ---
+    mapping_back = checkpoint.get("label_mapping", None)
+    if mapping_back is not None:
+        # convert possible tensor keys to int
+        mapping_back = {int(k): int(v) for k, v in mapping_back.items()}
+
+    # --- Auto-detect num_classes from checkpoint or mapping ---
     if num_classes is None:
-        num_classes = checkpoint.get("num_classes", None)
-        if num_classes is None:
-            if "model_state_dict" in checkpoint and "classifier.weight" in checkpoint["model_state_dict"]:
-                num_classes = checkpoint["model_state_dict"]["classifier.weight"].shape[0]
-            else:
-                num_classes = 2  # fallback
+        if "num_classes" in checkpoint and isinstance(checkpoint["num_classes"], int):
+            num_classes = checkpoint["num_classes"]
+        elif "model_state_dict" in checkpoint and "classifier.weight" in checkpoint["model_state_dict"]:
+            num_classes = checkpoint["model_state_dict"]["classifier.weight"].shape[0]
+        elif mapping_back is not None:
+            num_classes = len(mapping_back)
+        else:
+            num_classes = 2  # fallback
         print(f"🔍 Detected num_classes={num_classes} from checkpoint")
 
     # --- Build model skeleton ---
@@ -352,15 +531,13 @@ def classification_CLIP_finetuned(
         ckpt_head_dim = state_dict["classifier.weight"].shape[0]
         if ckpt_head_dim != num_classes:
             head_mismatch = True
-            print(f"⚠️ Head mismatch: checkpoint={ckpt_head_dim}, model={num_classes}. "
-                  f"Re-initializing classifier head ...")
-            # Remove incompatible head weights
+            print(f"⚠️ Head mismatch: checkpoint={ckpt_head_dim}, model={num_classes}. Re-initializing classifier head.")
             state_dict.pop("classifier.weight", None)
             state_dict.pop("classifier.bias", None)
 
     missing, unexpected = model.load_state_dict(state_dict, strict=False)
     if missing:
-        print(f"⚙️ Missing keys (expected for head reset): {missing}")
+        print(f"⚙️ Missing keys (ok if head was reset): {missing}")
     if unexpected:
         print(f"⚙️ Unexpected keys: {unexpected}")
     if head_mismatch:
@@ -368,7 +545,7 @@ def classification_CLIP_finetuned(
 
     model.eval()
 
-    # --- Dataset + Dataloader ---
+    # --- Dataset and DataLoader ---
     dataset = NewsDataset(
         df,
         processor,
@@ -389,183 +566,19 @@ def classification_CLIP_finetuned(
                 inputs[k] = v.to(_device)
             logits, _ = model(**inputs)
             _, predicted = logits.max(1)
-            predictions.extend((predicted + 1).cpu().numpy())
+
+            if mapping_back:
+                predictions.extend([mapping_back[int(p.cpu().item())] for p in predicted])
+            else:
+                # fallback - assume 1-based desired output
+                predictions.extend((predicted + 1).cpu().numpy())
 
     df[predict_column] = predictions
-    print(f"✅ Prediction complete! {len(df)} rows labeled.")
+    print(f"✅ Prediction complete - {len(df)} rows labeled with original class IDs.")
     return df
+# ==== PATCH: classification_CLIP_finetuned (end) =============================
 
 
-def finetune_CLIP(
-    mode="both",
-    text_path=None,
-    text_column=None,
-    img_dir=None,
-    true_label=None,
-    prompt=None,
-    model_name="best_clip_model.pth",
-    num_epochs=20,
-    batch_size=8,
-    learning_rate=1e-5
-):
-    """
-    Fine-tune CLIP on a text, image, or multimodal dataset.
-
-    Automatically detects the number of classes from the `true_label` column.
-    """
-
-    if mode not in ["text", "image", "both"]:
-        raise ValueError("mode must be one of 'text', 'image', or 'both'")
-
-    use_text  = mode in ["text", "both"]
-    use_image = mode in ["image", "both"]
-
-    if use_text and not text_path:
-        raise ValueError("text_path cannot be empty")
-    if use_image and not img_dir:
-        raise ValueError("img_dir cannot be empty")
-
-    # === Load dataset ===
-    if text_path.endswith(".csv") or text_path.endswith(".txt"):
-        df = pd.read_csv(text_path)
-    elif text_path.endswith(".jsonl"):
-        df = pd.read_json(text_path, lines=True)
-    elif text_path.endswith((".xlsx", ".xls")):
-        df = pd.read_excel(text_path)
-    else:
-        raise ValueError("Unsupported file format")
-
-    print(f"📂 Loaded {len(df)} records from {os.path.basename(text_path)}")
-
-    if true_label not in df.columns:
-        raise ValueError(f"Label column '{true_label}' not found in dataset")
-
-    # === Auto-detect number of classes ===
-    unique_labels = sorted(df[true_label].dropna().unique())
-    num_classes = len(unique_labels)
-
-    # Handle indexing (0 vs 1-based)
-    if 0 not in unique_labels and 1 in unique_labels:
-        label_min = int(min(unique_labels))
-        label_shift = 1 - label_min
-        if label_shift != 0:
-            df[true_label] = df[true_label] + label_shift
-    else:
-        label_shift = 0
-
-    print(f"🔍 Detected {num_classes} classes: {unique_labels}")
-    if label_shift != 0:
-        print(f"⚙️ Shifted labels by +{label_shift} for zero-based indexing.")
-
-    # === Train/validation split ===
-    if len(df) < 5:
-        train_df, val_df = df, df
-        print("⚠️ Dataset too small for validation split; using full dataset for training.")
-    else:
-        val_size = int(len(df) * 0.2)
-        train_df = df.iloc[:-val_size].reset_index(drop=True)
-        val_df   = df.iloc[-val_size:].reset_index(drop=True)
-
-    # === Device ===
-    training_device = device
-    if device.type == "mps":
-        training_device = torch.device("cpu")
-        print(f"⚠️ MPS detected → using CPU to avoid tensor layout issues.")
-    print(f"💻 Using device: {training_device}")
-
-    print(f"🧠 Training setup:")
-    print(f"   • Mode: {mode}")
-    print(f"   • Text columns: {text_column}")
-    print(f"   • Label column: {true_label}")
-    print(f"   • Number of classes: {num_classes}")
-    print(f"   • Batch size: {batch_size}, Epochs: {num_epochs}, LR: {learning_rate}")
-
-    if prompt:
-        print(f"   • Prompt: {prompt}")
-
-    # === Model & processor ===
-    clip_model = CLIPModel.from_pretrained(base_model)
-    processor  = CLIPProcessor.from_pretrained(base_model)
-
-    train_dataset = NewsDataset(
-        train_df, processor, text_column=text_column, img_dir=img_dir,
-        use_text=use_text, use_image=use_image, true_label=true_label, prompt=prompt
-    )
-    val_dataset = NewsDataset(
-        val_df, processor, text_column=text_column, img_dir=img_dir,
-        use_text=use_text, use_image=use_image, true_label=true_label, prompt=prompt
-    )
-
-    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
-    val_loader   = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)
-
-    # === Model, optimizer, loss ===
-    model = CLIPClassifier(clip_model, num_classes, use_text, use_image).to(training_device)
-    optimizer = AdamW(model.parameters(), lr=learning_rate)
-    criterion = torch.nn.CrossEntropyLoss()
-
-    # === Training loop ===
-    best_accuracy = 0.0
-    for epoch in range(num_epochs):
-        model.train()
-        total_loss, correct, total = 0, 0, 0
-
-        for batch_idx, (inputs, labels) in enumerate(tqdm(train_loader, desc=f"Epoch {epoch+1}/{num_epochs} [Train]")):
-            for k, v in inputs.items():
-                inputs[k] = v.to(training_device)
-            labels = labels.to(training_device)
-
-            optimizer.zero_grad()
-            logits, _ = model(**inputs)
-            loss = criterion(logits, labels)
-            loss.backward()
-            optimizer.step()
-
-            total_loss += loss.item()
-            _, predicted = logits.max(1)
-            total += labels.size(0)
-            correct += predicted.eq(labels).sum().item()
-
-        train_acc = 100. * correct / total
-        train_loss = total_loss / len(train_loader)
-
-        # === Validation ===
-        model.eval()
-        val_correct, val_total, val_loss = 0, 0, 0.0
-        with torch.no_grad():
-            for inputs, labels in tqdm(val_loader, desc=f"Epoch {epoch+1}/{num_epochs} [Val]"):
-                for k, v in inputs.items():
-                    inputs[k] = v.to(training_device)
-                labels = labels.to(training_device)
-
-                logits, _ = model(**inputs)
-                loss = criterion(logits, labels)
-                val_loss += loss.item()
-                _, predicted = logits.max(1)
-                val_total += labels.size(0)
-                val_correct += predicted.eq(labels).sum().item()
-
-        val_acc = 100. * val_correct / val_total if val_total else 0
-        val_loss = val_loss / len(val_loader)
-
-        print(f"Epoch {epoch+1}/{num_epochs}: "
-              f"Train Loss={train_loss:.4f} Acc={train_acc:.2f}% | "
-              f"Val Loss={val_loss:.4f} Acc={val_acc:.2f}%")
-
-        # === Save best model ===
-        if val_acc > best_accuracy:
-            best_accuracy = val_acc
-            torch.save({
-                "model_state_dict": model.state_dict(),
-                "epoch": epoch + 1,
-                "best_accuracy": best_accuracy,
-                "num_classes": num_classes,
-                "label_mapping": unique_labels
-            }, model_name)
-            print(f"✅ Model saved! New best validation accuracy: {best_accuracy:.2f}%")
-
-    print(f"🎯 Fine-tuning complete! Best validation accuracy: {best_accuracy:.2f}%")
-    return best_accuracy
 
 # =========================
 # Shared utilities
@@ -627,13 +640,17 @@ class ClassificationQuestion:
     max_verify_retry: int = 2
 
     def get_key(self) -> str:
+        # Normalize text & prompt to improve cache hit chance
+        norm_prompt = (self.prompt or "").strip().lower()
+        norm_text = (self.text or "").strip().lower()
+    
         parts = [
-            self.prompt,
+            norm_prompt,
             self.model_name,
-            ",".join(self.valid_values),
+            ",".join(sorted(self.valid_values)),  # sort to normalize order
             str(self.reasoning_effort),
             self.column_4_labeling,
-            self.text or "",
+            norm_text,
             str(self.label_num),
             str(self.max_verify_retry),
         ]
@@ -649,7 +666,7 @@ class ClassificationTask:
     column_4_labeling: str
     label_num: int = 1
     once_verify_num: int = 1
-    max_verify_retry: int = 3
+    max_verify_retry: int = 5
 
     def create_question(self, content: str) -> ClassificationQuestion:
         return ClassificationQuestion(
@@ -684,6 +701,46 @@ class GPTClassifier:
         self.client = client
         self.cache  = DBCache()
 
+    @staticmethod
+    def _validate_output(candidate, valid_values: list[str], num_themes: int):
+        """
+        Normalize and validate a single candidate output.
+
+        Returns:
+            list[int] of length num_themes if valid, else None.
+        """
+        if candidate is None:
+            return None
+
+        # Normalize candidate to a list
+        if isinstance(candidate, (str, int)):
+            candidate = [candidate]
+
+        if not isinstance(candidate, (list, tuple)):
+            return None
+
+        # Convert values to strings, strip, and check membership
+        norm_strs = []
+        for v in candidate:
+            s = str(v).strip()
+            # accept bare integers like 5 as "5"
+            if s.isdigit() and s in valid_values:
+                norm_strs.append(s)
+            else:
+                # if it looks like "05" or non-digit tokens, reject
+                return None
+
+        # Length must match exactly
+        if len(norm_strs) != int(num_themes):
+            return None
+
+        # All values must be in valid_values
+        if any(s not in valid_values for s in norm_strs):
+            return None
+
+        # Convert to ints for downstream convenience
+        return [int(s) for s in norm_strs]
+
     # -- Helpers: Responses-API wrapping --
     @staticmethod
     def _to_responses_input(messages):
@@ -771,6 +828,12 @@ class GPTClassifier:
 
     # -- SINGLE call → parsed labels (now forwards temperature) --
     def classify(self, q: ClassificationQuestion, n: int, temperature: float | None = None):
+        """
+        Run a single GPT call that may return up to n choices.
+        Returns:
+            parsed: list of parsed candidates (each candidate is list or scalar-like)
+            raw_texts: list of raw text replies for logging
+        """
         if q.column_4_labeling == "text_class":
             content = [
                 {"type": "text", "text": str(q.prompt)},
@@ -801,58 +864,229 @@ class GPTClassifier:
         )
 
         parsed = []
-        for reply in (c.message.content.strip() for c in resp.choices):
-            if reply.startswith("[") and reply.endswith("]"):
+        raw_texts = []
+        for choice in resp.choices:
+            raw = choice.message.content.strip() if getattr(choice, "message", None) else ""
+            raw_texts.append(raw)
+
+            # Try JSON-like list first
+            if raw.startswith("[") and raw.endswith("]"):
                 try:
-                    arr = json.loads(reply)
-                    if isinstance(arr, list) and len(arr) == q.label_num and all(str(x) in q.valid_values for x in arr):
-                        parsed.append([int(x) for x in arr])
-                        continue
+                    arr = json.loads(raw)
+                    parsed.append(arr)
+                    continue
                 except Exception:
                     pass
-            flat = re.findall(r"\b[01]\b", reply)
-            if len(flat) == q.label_num:
-                parsed.append([int(x) for x in flat])
+
+            # Try simple digits separated by commas or spaces
+            flat = re.findall(r"\b\d+\b", raw)
+            if flat:
+                parsed.append([int(x) if x.isdigit() else x for x in flat])
                 continue
-            matches = [p.strip(' ."') for p in reply.split(",") if p.strip(' ."') in q.valid_values]
-            if matches:
-                parsed.append(matches[: q.label_num])
-                continue
-            m = re.search(r"\b(\d{1,2})\b", reply)
-            if m and m.group(1) in q.valid_values:
-                parsed.append([m.group(1)])
+
+            # As a last resort, keep raw as-is (will fail validation)
+            parsed.append(raw)
 
         if not parsed:
             logger.error(f"No valid labels parsed. Raw reply: {resp.choices}")
-        return parsed
+        return parsed, raw_texts
+
 
     # -- majority vote / cache (now forwards temperature) --
-    def multi_verify(self, q: ClassificationQuestion, n, retry=1, freq=None, temperature: float | None = None):
-        res = self.classify(q, n, temperature=temperature)
-        if not res and retry < q.max_verify_retry:
-            return self.multi_verify(q, n, retry + 1, temperature=temperature)
-        if not res:
-            raise MaxRetryException("No valid responses")
+    def multi_verify(
+        self,
+        q: ClassificationQuestion,
+        n,
+        retry=1,
+        freq=None,
+        temperature: float | None = None,
+        counters: dict | None = None
+    ):
+        """
+        Robust per-row classification:
+          - Checks SQLite cache first
+          - Validates each attempt immediately
+          - Retries up to q.max_verify_retry times
+          - Caches only valid outputs
+          - Returns [99,...] on final failure (not cached)
+          - IMPORTANT: counters['redo'] counts TRUE retries only
+                       (i.e., second attempt and beyond), never the first try.
+        """
+        # 1) Cache check
+        cached = self.cache.get(q)
+        if cached is not None:
+            return cached
+    
+        max_retry = max(1, int(q.max_verify_retry))
+        attempt = 1
+        total_retries_this_row = 0
+        last_error = None
+        row_tag = f"{getattr(q, 'row_idx', '?')}"
+    
+        while attempt <= max_retry:
+            try:
+                parsed_list, raw_list = self.classify(q, n, temperature=temperature)
+            except Exception as e:
+                last_error = e
+                # Only a retry if we're going to try again
+                if attempt < max_retry:
+                    total_retries_this_row += 1
+                    if counters is not None:
+                        counters["redo"] += 1
+                    print(f"Row {row_tag} RETRY #{total_retries_this_row} — API error: {type(e).__name__}: {e}")
+                    attempt += 1
+                    continue
+                # No more retries left
+                break
 
-        if isinstance(res[0], list) and len(res[0]) == q.label_num:
-            self.cache.add(q, res[0])
-            return res[0]
+            # Validate candidates
+            counts = {}
+            first_parsed_preview = None
+            first_raw_preview = None
+    
+            for idx, cand in enumerate(parsed_list or []):
+                if first_parsed_preview is None:
+                    first_parsed_preview = cand
+                if first_raw_preview is None:
+                    first_raw_preview = (raw_list[idx] if raw_list and idx < len(raw_list) else "")
+    
+                valid = self._validate_output(cand, q.valid_values, q.label_num)
+                if valid is not None:
+                    t = tuple(valid)
+                    counts[t] = counts.get(t, 0) + 1
+    
+            if counts:
+                # Success: majority vote
+                best_tuple = sorted(counts.items(), key=lambda kv: kv[1], reverse=True)[0][0]
+                best = list(best_tuple)
+                self.cache.add(q, best)
+                return best
+    
+            # Invalid output
+            parsed_preview_str = str(first_parsed_preview)
+            raw_preview_str = (first_raw_preview or "").replace("\n", " ")[:200]
+    
+            if attempt < max_retry:
+                # We WILL retry → count it
+                total_retries_this_row += 1
+                if counters is not None:
+                    counters["redo"] += 1
+                print(
+                    f"Row {row_tag} RETRY #{total_retries_this_row} — invalid output: "
+                    f"{parsed_preview_str} (expected {q.label_num} labels). Raw: \"{raw_preview_str}\""
+                )
+                attempt += 1
+                continue
+    
+            # No retries left, fall through to failure
+            break
+    
+        # Final failure
+        fail_reason = (
+            f"{type(last_error).__name__}: {last_error}"
+            if last_error is not None else
+            "invalid format repeatedly."
+        )
+        print(f"Row {row_tag} FAILED after {total_retries_this_row} retries. Last reason: {fail_reason}")
+        return [99] * int(q.label_num)
+
+    
+    def classify_df(
+        self,
+        df: pd.DataFrame,
+        task: ClassificationTask,
+        return_sample_response=False,
+        temperature: float | None = None,
+        pbar=None,
+        counters=None
+    ):
+        """
+        Row-by-row processing:
+          - Counts cache hits in counters['cache']
+          - Counts TRUE retries in counters['redo'] (handled inside multi_verify)
+          - Progress bar postfix shows: ⚡ cache:<n> | retry:<m>
+        """
+        out, sample = [], None
+    
+        for idx, rec in enumerate(df.to_dict("records")):
+            q = task.create_question(rec.get(task.column, ""))
+            setattr(q, "row_idx", idx)
+    
+            # Cache check BEFORE classification (does not affect redo)
+            if self.cache.get(q) is not None:
+                if counters is not None:
+                    counters["cache"] += 1
+    
+            if return_sample_response and sample is None:
+                sample = self.fetch(
+                    [{"role": "user",
+                      "content": [
+                          {"type": "text", "text": str(q.prompt)},
+                          {"type": "text", "text": str(q.text)}
+                      ]}],
+                    q.model_name,
+                    q.reasoning_effort or "minimal",
+                    1,
+                    temperature=temperature,
+                )
+    
+            try:
+                lbl = self.multi_verify(
+                    q,
+                    task.once_verify_num,
+                    temperature=temperature,
+                    counters=counters,  # <- redo counted ONLY here, on true retries
+                )
+            except Exception as e:
+                print(f"Row {idx} ERROR — {type(e).__name__}: {e}")
+                lbl = [99] * int(task.label_num)
+    
+            rec[task.column_4_labeling] = lbl
+            out.append(rec)
+    
+            # Progress bar update
+            if pbar is not None:
+                postfix = (
+                    f"⚡ cache:{counters['cache']}"
+                    if counters else ""
+                )
+                pbar.set_postfix_str(postfix, refresh=True)
+                pbar.update(1)
+    
+        df_out = pd.DataFrame(out)
+        return (df_out, sample) if return_sample_response else df_out
 
-        freq = freq or {}
-        for r in res:
-            lbl = r[0]
-            freq[lbl] = freq.get(lbl, 0) + 1
-        top = sorted(freq, key=freq.get, reverse=True)[: q.label_num]
-        self.cache.add(q, top)
-        return top
 
     # -- DataFrame helper (now forwards temperature) --
     def classify_df(self, df: pd.DataFrame, task: ClassificationTask,
-                    return_sample_response=False, temperature: float | None = None):
+                    return_sample_response=False, temperature: float | None = None,
+                    pbar=None, counters=None):
+        """
+        Processes dataframe row by row:
+        ✅ Tracks cache hits & re-runs
+        ✅ Shows ⚡ cache inline on progress bar
+        ✅ Updates tqdm postfix dynamically
+        """
         out, sample = [], None
-        for rec in df.to_dict("records"):  # no tqdm here
+    
+        for idx, rec in enumerate(df.to_dict("records")):
             q = task.create_question(rec.get(task.column, ""))
+            try:
+                setattr(q, "row_idx", idx)
+            except:
+                pass
+    
+            # Check cache BEFORE calling multi_verify
+            cache_hit = self.cache.get(q) is not None
+    
+            if cache_hit:
+                if counters:
+                    counters["cache"] += 1
+            else:
+                if counters:
+                    counters["redo"] += 1
 
+    
             if return_sample_response and sample is None:
                 sample = self.fetch(
                     [{"role": "user",
@@ -863,16 +1097,23 @@ class GPTClassifier:
                     1,
                     temperature=temperature,
                 )
-
+    
             try:
                 lbl = self.multi_verify(q, task.once_verify_num, temperature=temperature)
             except Exception as e:
-                logger.error(e)
-                lbl = ["99"] * task.label_num
+                # Print clear reason for retry
+                print(f"Row {getattr(q, 'row_idx', '?')} RETRY #{attempts+1} — error: {str(e)}")
+                lbl = [99] * int(q.label_num)
 
             rec[task.column_4_labeling] = lbl
             out.append(rec)
-
+    
+            # ✅ tqdm updates
+            if pbar is not None:
+                postfix = f"⚡ cache: {counters['cache']} " if counters else ""
+                pbar.set_postfix_str(postfix, refresh=True)
+                pbar.update(1)
+    
         df_out = pd.DataFrame(out)
         return (df_out, sample) if return_sample_response else df_out
 
@@ -982,9 +1223,23 @@ def classification_GPT(
     outputs = []
     n = len(df)
     
-    # one global progress bar
-    pbar = tqdm(total=n, desc=f"Classifying {tasks[0][1]}", unit="item")
+    # Global counters
+    counters = {"cache": 0, "redo": 0}
     
+    pbar = tqdm(
+        total=n,
+        desc=f"[ GPT • {tasks[0][1]} ]",
+        unit="row",
+        ncols=100,
+        dynamic_ncols=False,
+        leave=True,
+        position=0,
+        mininterval=0.5,
+        smoothing=0.1,
+        bar_format="{desc} {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} | ETA {remaining} | {rate_fmt} | {postfix}"
+    )
+
+
     for start in range(0, n, max(1, batch_size)):
         end = min(n, start + max(1, batch_size))
         sub = df.iloc[start:end].copy()
@@ -994,17 +1249,24 @@ def classification_GPT(
                 column=col, prompt=pr, model_name=model,
                 valid_values=category, reasoning_effort=reasoning_effort,
                 column_4_labeling=lab, label_num=num_themes,
-                once_verify_num=num_votes, max_verify_retry=num_votes,
+                once_verify_num=num_votes, max_verify_retry=5,
             )
     
-            # tell classify_df not to create its own tqdm
-            sub = clf.classify_df(sub, task, temperature=temperature)
-        
+            # Pass counters + tqdm bar
+            sub = clf.classify_df(sub, task, temperature=temperature, pbar=pbar, counters=counters)
+    
         outputs.append(sub)
-        pbar.update(len(sub))  # move main progress bar
         time.sleep(wait_time)
     
+    # Ensure bar completes cleanly
+    pbar.n = pbar.total
+    pbar.refresh()
     pbar.close()
+    
+    print(f"\n✅ Finished classification of {n} rows.")
+
+
+    
 
 
     df = pd.concat(outputs, ignore_index=True)
@@ -1224,4 +1486,4 @@ def auto_verification(
     overall_accuracy = total_correct / total_count if total_count else 0.0
     print(f"\n>> Overall accuracy: {overall_accuracy:.2%}")
     overall_results["overall_accuracy"] = overall_accuracy
-    return overall_results
+    return overall_results
\ No newline at end of file
